/* -*- mode: C -*-

  scsi_vpd_render - Render SCSI VPD components using templates.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: scsi_vpd_render.c,v 1.16 2004/10/14 00:39:46 martins Exp $
*/
  
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <regex.h>
#include <ctype.h>
#include <unistd.h>

#include "util.h"

#define TEMPLATE_FILE LSVPD_LIBDIR "/scsivpd.conf"

static void
usage()
{
	fprintf(stderr, "usage: %s indir outdir\n", vpd_program_name);
	exit(1);
}

static char *
hexify(const char * s, size_t len)
{
	int i;
	char * ret = malloc(len*2+1);
	for (i = 0; i < len ; i++) {
		snprintf(&ret[i*2], 3, "%.2X", (unsigned char) s[i]);
	}

	return ret;
}

static void
strstrip(char * s)
{
	char *b, *t, *e;

	b = s;
	while (isspace(*b)) {
		b++;
	}

	e = t = b;
	while (isprint(*t)) {
		if (!isspace(*t)) {
			e = t;
		}
		t++;
	}
	e++;

	memmove(s, b, e - b);
	s[e - b] = '\0';
}

static void
dump_field(const char * dir,
	   const char * key,
	   const char * data,
	   size_t       len,
	   const char * dtype)
{
	/*
	 * Many fields will be whitespace-stripped below, so we'll
	 * need to make sure the string is NUL-terminated.  This means
	 * we need to take a copy of the string.
	 */
	char * ldata;

	/* Certain fields will require some munging. */
	/* Convert RL, Z0, Z7 to hex. */
	if (((strcmp(key, "RL") == 0) && (strcmp(dtype, "disk") == 0)) ||
	    (strcmp(key, "Z0") == 0) ||
	    (strcmp(key, "Z7") == 0)) {
		/* hexify returns a new NUL-terminated string. */
		ldata = hexify(data, len);
	} else {
		/* Copy data and NUL-terminate if necessary. */
		ldata = malloc(len+1);
		memmove(ldata, data, len);
		ldata[len] = '\0';
	}

	/* This also terminates at the first NUL. */
	strstrip(ldata);

	/* Write data to file. */
	file_relative_write_contents(dir, key, ldata, strlen(ldata));

	free(ldata);
}

#define ARRAY_LEN(x) (sizeof((x))/sizeof((x)[0]))

/* Very vaguely taken from linux-2.5.22/drivers/scsi/scsi.[ch]
 * There are only really 14 possible type, but filling it out to 16
 * removes the need for error checking. */
static const struct string_array scsi_device_types_short[] =
{
        {0x0, "disk"},
        {0x1, "tape"},
        {0x2, "printer"},
        {0x3, "processor"},
        {0x4, "worm"},
        {0x5, "cdrom"},
        {0x6, "scanner"},
        {0x7, "optical-disk"},
        {0x8, "jukebox"},
        {0x9, "communications"},
        {0xa, "graphics"},
        {0xb, "graphics"},
        {0xd, "enclosure"},
        {-1, NULL}
};


static const struct string_array scsi_device_types[] =
{
        {0x0, "Disk Drive"},
        {0x1, "Tape Drive"},
        {0x2, "Printer Device"},
        {0x3, "Processor Device"},
        {0x4, "Write-once Device"},
        {0x5, "CD-ROM Drive"},
        {0x6, "Scanner Device"},
        {0x7, "Optical Memory Device"},
        {0x8, "Medium Changer Device"},
        {0x9, "Communications Device"},
        {0xa, "Graphics Arts Pre-press Device"},
        {0xb, "Graphics Arts Pre-press Device"},
        {0xd, "Enclosure Services Device"},
        {-1, NULL}
};

#define SCSI_CDROM_DEFAULT "Other SCSI CD-ROM Drive"
static const struct string_array cdrom_device_types[] =
{
	{0, SCSI_CDROM_DEFAULT},
	{1, "OEM SCSI CD-ROM Drive"},
	{2, "SCSI CD-ROM Drive"},
	{3, "SCSI Multimedia CD-ROM Drive"},
	{4, "OEM SCSI DVD-ROM Drive"},
	{5, "SCSI DVD-ROM Drive"},
	{6, "SCSI DVD-RAM Drive"},
	{7, "OEM SCSI DVD-RAM Drive"},
	{-1, NULL}
};

static char *
make_ds_value(const char * data, unsigned char dtype, const char * dir)
{
	char * data_0xc7 = NULL;
	size_t len_0xc7;
	int err;
	char  * file = NULL;

	char *s, *ret;

	/* Get info for naming quirks, such as LVD, Multimedia. */
	file = build_relative_filename(dir, "0xc7");
	file_read_contents(file, &data_0xc7, &len_0xc7, &err);
	free(file);

	/* Now do DS and type. */

	/* Check for SCSI *Multimedia* CD-ROM Drive. */
	if (dtype == 5) {
		int type = 0;
		if ((data_0xc7 != NULL) && (len_0xc7 >=34)) {
			type = (int) data_0xc7[33];
		}
		/* So it can be freed later. */
		asprintf(&s, "%s",
			 string_array_lookup(cdrom_device_types,	
					     SCSI_CDROM_DEFAULT,
					     type));
	} else {
		if ((dtype == 0) &&
		    (data_0xc7 != NULL) &&
		    (len_0xc7 >=35) && (data_0xc7[34] == 7)) {
			/* *LVD* SCSI Disk Drive. */
			asprintf(&s,
				 "LVD SCSI %s",
				 string_array_lookup (scsi_device_types,
						      "Unknown Device",
						      dtype));
		} else {
			asprintf(&s,
				 "SCSI %s",
				 string_array_lookup(scsi_device_types,
						     "Unknown Device",
						     dtype));
		}
	}

	asprintf(&ret, "%s%s",
		 ((data[7] & 0x40) ? "32 Bit " :
		  ((data[7] & 0x20) ? "16 Bit " : "")),
		 s);

	free(s);

	if (NULL != data_0xc7) {
		free(data_0xc7);
	}
	
	return ret;
}

static char *
get_templates(const char * vendor,
	      const char * type,
	      const char * model)
{
	char * buf = NULL;
	size_t length;
	int    err;

	char * s;
	char * t;
	char * max;

	char * vtm;
	regex_t preg;

	char * ret = NULL;

	if (!file_read_contents(TEMPLATE_FILE, &buf, &length, &err)) {
		die2(strerror(err), TEMPLATE_FILE);
	}
	max = buf + length;

	asprintf(&vtm, "%s;%s;%s", vendor, type, model);

	/* Search for template. */
	s = buf;
	while ((NULL == ret) && (s < max)) {
		int i;
		char * templates;

		/* Comment? */
		if ('#' == *s) {
			if (NULL == (s = memchr(s, '\n', max - s))) {
				s = max;
			} else {
				s++;
			}
			continue;
		}
				
		/* Pattern starts at beginning of line. */
		t = s;

		/* Find the 3rd semicolon, remember, NUL-terminate. */
		i = 0;
		while ((i < 3) && (NULL != (s = memchr(s, ';', max - s)))) {
			i++;
			s++;
		}

		templates = NULL;

		if (3 == i) {
			*(s-1) = '\0';

			/* If match this template, remember this spot. */
			if (0 == regcomp(&preg, t, REG_EXTENDED|REG_NOSUB)) {
				if (0 == regexec(&preg, vtm, 0, NULL, 0)) {
					templates = s;
				}
				regfree(&preg);
			} else {
				fprintf(stderr,
					"%s: failed to compile regexp \"%s\"\n",
					vpd_program_name, t);
			}
		}

		/* Find newline and terminate there. */
		if (NULL == (s = memchr(s, '\n', max - s))) {
			/* There's a NUL-terminator there.  Let's use it! */
			s = max + 1;
		} else {
			*s = '\0';
			s++;
		}

		/* OK got it - move to front of buf. */
		if (NULL != templates) {
			memmove(buf, templates, (size_t) (s - templates));
			ret = buf;
		}
	}

	free(vtm);

	return ret;
}

static char *
terminate_or_maybe_die(char * s, char c, char * max, bool die_if_fail)
{
	if (NULL == (s = memchr(s, c, max - s))) {
		if (die_if_fail) {
			/* Bogus template. */
			die("Internal error - invalid template");
		} else {
			/* Allowed to hit end of template. NUL there! */
			s = max;
		}
	} else {
		*s = '\0';
		s++;
	}
	return s;
}	

static void
process_template(char * template,
		 const char * indir,
		 const char * outdir,
		 const char * type)
{
	char * buf = NULL;
	size_t length;
	int err;
	int offset;

	char * file;

	char * s;
	char * tmax;

	tmax = template + strlen(template);

	/* NUL-terminate page name. */
	s = template;
	s = terminate_or_maybe_die(s, '=', tmax, true);

	/* Can't read file?  No problem: template might be too specific. */
	file = build_relative_filename(indir, template);
	if (!file_read_contents(file, &buf, &length, &err)) {
		free(file);
		return;
	}
	free(file);

	if (strcmp(template, "inquiry") == 0) {
		offset = 32;
	} else {
		offset = 0;
	}

	while ((s < tmax) && (offset < length)) {
		char *k, *v;
		int field_len;

		k = s;
		s = terminate_or_maybe_die(s, ':', tmax, true);
		v = s;
		s = terminate_or_maybe_die(s, ',', tmax, false);
		
		field_len = atoi(v);

		if ((k[0] != '_') &&
		    (offset + field_len <= length)){
			dump_field(outdir, k, buf + offset,  field_len, type);
		}
		offset += field_len;
	}

	free(buf);
}

static void
process_templates(char * templates,
		  const char * indir,
		  const char * outdir,
		  const char * type)
{
	char * s = templates;
	char * tmax = templates + strlen(templates);

	while (s < tmax) {
		char * t = s;

		s = terminate_or_maybe_die(s, ';', tmax, false);
		process_template(t, indir, outdir, type);
	}
}

int
main(int argc, char **argv)
{
	char * mf = NULL;
	char * tm  = NULL;
	size_t len;

	char * data = NULL;

	unsigned char dtype;
	const char * dtype_short;
	char * dtype_str;

	char * templates;
	
	vpd_program_name = argv[0];

	if (argc != 3) {
		usage();
	}

	/* Grab contents of inquiry file. */
	file_relative_read_contents(argv[1], "inquiry", &data, &len);

	/* Validate data. */
	if (len < 32) {
		fprintf(stderr, "%s: short standard inquiry data in %s/%s\n",
			vpd_program_name, argv[1], "inquiry");
		exit(1);
	}

	dtype = data[0] & 0x1F;
	dtype_short = string_array_lookup(scsi_device_types_short,
					  "unknown",
					  dtype);

	/* Parse out standard fields. */
	dump_field(argv[2], "Z0", data,    8,  dtype_short);

	asprintf(&mf, "%.*s",  8, data +  8);
	strstrip(mf);
	dump_field(argv[2], "MF", mf,  8,  dtype_short);

	asprintf(&tm, "%.*s", 16, data + 16);
	strstrip(tm);
	dump_field(argv[2], "TM", tm, 16, dtype_short);

	/* Now do DS and type. */
	dtype_str = make_ds_value(data, dtype, argv[1]);
	dump_field(argv[2], "DS", dtype_str, strlen(dtype_str), dtype_short);
	free(dtype_str);

	free(data);

	/* "type" is used externally. */
	file_relative_write_contents(argv[1], "type",
				     dtype_short, strlen(dtype_short));

	/* Parse out extra fields using templates. */
	templates = get_templates(mf, dtype_short, tm);
	if (templates) {
		process_templates(templates, argv[1], argv[2], dtype_short);
		free(templates);
	}

	free(mf);
	free(tm);

	return 0;
}
